import sys

from loguru import logger


def configure_logger(verbose: bool, format_: str):
    logger.remove()

    if format_ == "porcelain":
        logger.add(sys.stderr, level="CRITICAL")
        return
    else:
        logger.add(
            sys.stderr,
            level="DEBUG" if verbose else "INFO",
            colorize=True,
            format="<level>{message}</level>",
        )


def print_results(deps_resolver, format_: str):
    unused_packages = deps_resolver.get_unused_package_names()
    if unused_packages:
        if format_ == "porcelain":
            print("\n".join(unused_packages))
        else:
            logger.error(f"Unused packages found: {', '.join(unused_packages)}")
        # sys.exit(1)
    else:
        logger.info("No unused packages found.")



import argparse
import glob
import sys

from loguru import logger

from creosote import formatters, parsers, resolvers


def parse_args(args):
    parser = argparse.ArgumentParser(
        description=(
            "Prevent bloated virtual environments by identifing installed, "
            "but unused, dependencies"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--verbose",
        dest="verbose",
        action="store_true",
        help="increase output verbosity",
    )
    parser.add_argument(
        "-f",
        "--format",
        dest="format",
        default="default",
        choices=["default", "porcelain"],
        help="output format",
    )
    parser.add_argument(
        "-p",
        "--paths",
        dest="paths",
        default=glob.glob("src"),
        nargs="*",
        help=("one or more paths to Python files"),
    )
    parser.add_argument(
        "-v",
        "--venv",
        dest="venv",
        default=".venv",
        help="path to the virtual environment you want to scan",
    )

    parser.add_argument(
        "-d",
        "--deps-file",
        dest="deps_file",
        metavar="PATH",
        default="pyproject.toml",
        help="path to the pyproject.toml, *.txt or *.in dependencies file",
    )

    parser.add_argument(
        "--dev",
        dest="dev",
        action="store_true",
        help="scan dev dependencies instead of prod dependencies",
    )

    parsed_args = parser.parse_args(args)

    if "pyproject.toml" not in parsed_args.deps_file and parsed_args.dev:
        raise Exception("Option --dev must be used with pyproject.toml")

    return parsed_args


def main(args_=None):
    args = parse_args(args_)

    formatters.configure_logger(verbose=args.verbose, format_=args.format)

    imports = parsers.get_modules_from_code(args.paths)
    logger.debug("Imports found:")
    for imp in imports:
        logger.debug(imp)

    logger.info(f"Parsing {args.deps_file} for packages")
    deps_reader = parsers.PackageReader()
    deps_reader.read(args.deps_file, args.dev)

    logger.info("Resolving...")
    deps_resolver = resolvers.DepsResolver(
        imports=imports, packages=deps_reader.packages or [], venv=args.venv
    )
    deps_resolver.resolve()

    logger.debug("Packages:")
    for package in deps_resolver.packages:
        logger.debug(package)

    formatters.print_results(deps_resolver=deps_resolver, format_=args.format)


if __name__ == "__main__":
    main(sys.argv[1:])



import ast
import pathlib
from functools import lru_cache

import toml
from loguru import logger

from creosote.models import Import, Package


class PackageReader:
    def __init__(self):
        self.packages = None

    def _pyproject(self, deps_file: str, dev: bool):
        """Return dependencies from pyproject.toml."""
        with open(deps_file, "r") as infile:
            contents = toml.loads(infile.read())

        try:
            if dev:
                deps = contents["tool"]["poetry"]["dev-dependencies"]
            else:
                deps = contents["tool"]["poetry"]["dependencies"]
        except KeyError as e:
            raise Exception("Could not find expected toml property.") from e

        return sorted(deps.keys())

    def _requirements(self, deps_file: str):
        """Return dependencies from requirements.txt-format file."""
        deps = []
        with open(deps_file, "r") as infile:
            contents = infile.readlines()

        for line in contents:
            if not line.startswith(" "):
                deps.append(line[: line.find("=")])

        return sorted(deps)

    @lru_cache(maxsize=None)  # noqa: B019
    def ignore_packages(self):
        return ["python"]

    def packages_sans_ignored(self, deps):
        packages = []
        for dep in deps:
            if dep not in self.ignore_packages():
                packages.append(Package(name=dep))
        return packages

    def read(self, deps_file: str, dev: bool):
        if not pathlib.Path(deps_file).exists():
            raise Exception(f"File {deps_file} does not exist")

        if "pyproject.toml" in deps_file:
            self.packages = self.packages_sans_ignored(self._pyproject(deps_file, dev))
        elif deps_file.endswith(".txt") or deps_file.endswith(".in"):
            self.packages = self.packages_sans_ignored(self._requirements(deps_file))
        else:
            raise NotImplementedError(
                f"Dependency specs file {deps_file} is not supported."
            )

        logger.info(
            f"Found packages in {deps_file}: "
            f"{', '.join([pkg.name for pkg in self.packages])}"
        )


def get_module_info_from_code(path):
    """Get imports, based on given filepath.

    Credit:
        https://stackoverflow.com/a/9049549/2448495
    """
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    for node in ast.iter_child_nodes(root):  # or potentially ast.walk ?
        if isinstance(node, ast.Import):
            module = []
        elif isinstance(node, ast.ImportFrom):
            module = node.module.split(".")
        else:
            continue

        for n in node.names:
            yield Import(module, n.name.split("."), n.asname)


def get_modules_from_code(paths):
    resolved_paths = []
    imports = []

    for path in paths:
        if pathlib.Path(path).is_dir():
            resolved_paths.extend(iter(pathlib.Path(path).glob("**/*.py")))
        else:
            resolved_paths.append(pathlib.Path(path).resolve())

    for resolved_path in resolved_paths:
        logger.info(f"Parsing {resolved_path}")
        for imp in get_module_info_from_code(resolved_path):
            imports.append(imp)

    dupes_removed = []
    for imp in imports:
        if imp not in dupes_removed:
            dupes_removed.append(imp)

    return dupes_removed



import dataclasses
from typing import List, Optional


@dataclasses.dataclass
class Import:
    module: List[str]
    name: List[str]
    alias: Optional[str] = None


@dataclasses.dataclass
class Package:
    name: str
    top_level_names: Optional[List[str]] = None
    module_name: Optional[str] = None
    importable_as: Optional[str] = None
    associated_imports: List[Import] = dataclasses.field(default_factory=list)



import os
import pathlib
from typing import List, Optional

from distlib import database

from creosote.models import Import, Package


class DepsResolver:
    def __init__(
        self,
        imports: List[Import],
        packages: List[Package],
        venv: str,
    ):
        self.imports = imports
        self.packages = packages
        self.venv = venv

        self.unused_packages: Optional[List[Package]] = None

    @staticmethod
    def canonicalize_module_name(module_name: str):
        return module_name.replace("-", "_").replace(".", "_").strip()

    def is_importable(self, module_name: str):
        try:
            __import__(self.canonicalize_module_name(module_name))
            return True
        except ImportError:
            return False

    def top_level_names(self, package):
        package_name = package.name.replace("-", "_")
        site_path = pathlib.Path(".")
        glob_str = f"{self.venv}/**/{package_name}*.dist-info/top_level.txt"
        top_levels = site_path.glob(glob_str)
        for top_level in top_levels:
            with open(top_level, "r") as infile:
                lines = infile.readlines()
            package.top_level_names = [line.strip() for line in lines]

    def package_to_module(self, package: Package):
        dp = database.DistributionPath(include_egg=True)
        dist = dp.get_distribution(package.name)
        if dist is None:
            # raise ModuleNotFoundError
            return
        module = package.name  # until we figure out something better
        for filename, _, _ in dist.list_installed_files():
            if filename.endswith((".py")):
                parts = os.path.splitext(filename)[0].split(os.sep)
                if len(parts) == 1:  # windows sep varies with distribution type
                    parts = os.path.splitext(filename)[0].split("/")
                if parts[-1].startswith("_") and not parts[-1].startswith("__"):
                    continue  # ignore internals
                elif filename.endswith(".py") and parts[-1] == "__init__":
                    module = parts[-2]
                    break

        package.module_name = module

    def associate_imports_with_package(self, package: Package, name: str):
        for imp in self.imports.copy():
            if not imp.module and name in imp.name:
                # import <imp.name>
                package.associated_imports.append(imp)
            elif imp.name and name in imp.module:
                # from <imp.name> import ...
                package.associated_imports.append(imp)

    def populate_packages(self):
        for package in self.packages:
            self.top_level_names(package)
            self.package_to_module(package)

    def associate(self):
        for package in self.packages:
            self.associate_imports_with_package(package, package.name)
            if package.top_level_names:
                for t in package.top_level_names:
                    self.associate_imports_with_package(package, t)
            if package.module_name:
                self.associate_imports_with_package(package, package.module_name)

    def get_unused_packages(self):
        self.unused_packages = [
            package for package in self.packages if not package.associated_imports
        ]

    def get_unused_package_names(self):
        if self.unused_packages:
            return [package.name for package in self.unused_packages]
        return None

    def resolve(self):
        self.populate_packages()
        self.associate()
        self.get_unused_packages()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['creosote']

package_data = \
{'': ['*']}

install_requires = \
['distlib>=0.3.4,<0.4.0', 'loguru>=0.6.0,<0.7.0', 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['creosote = creosote.cli:main']}

setup_kwargs = {
    'name': 'creosote',
    'version': '1.0.2',
    'description': 'Identify unused dependencies and avoid a bloated virtual environment.',
    'long_description': '# creosote\n\n[![check](https://github.com/fredrikaverpil/creosote/actions/workflows/check.yaml/badge.svg)](https://github.com/fredrikaverpil/creosote/actions/workflows/check.yaml) [![test](https://github.com/fredrikaverpil/creosote/actions/workflows/test.yaml/badge.svg)](https://github.com/fredrikaverpil/creosote/actions/workflows/test.yaml)\n\nIdentify unused dependencies and avoid a bloated virtual environment.\n\n## Quickstart\n\nInstall creosote in separate virtual environment (using e.g. [pipx](https://github.com/pypa/pipx)):\n\n```bash\npipx install creosote\n```\n\nScan virtual environment for unused packages:\n\n```bash\ncreosote --deps-file pyproject.toml --venv .venv --paths src\n```\n\nExample output:\n\n```bash\n$ creosote\nParsing src/creosote/formatters.py\nParsing src/creosote/models.py\nParsing src/creosote/resolvers.py\nParsing src/creosote/__init__.py\nParsing src/creosote/parsers.py\nParsing src/creosote/cli.py\nParsing pyproject.toml for packages\nFound packages in pyproject.toml: PyYAML, distlib, loguru, protobuf, toml\nResolving...\nUnused packages found: PyYAML, protobuf\n```\n\nGet help:\n\n```bash\ncreosote --help\n```\n\n## How this works\n\nSome data is required as input:\n\n- A list of package names (fetched from e.g. `pyproject.toml`, `requirements_*.txt|.in`).\n- The path to the virtual environment.\n- The path to one or more Python files (or a folder containing such files).\n\nThe creosote tool will first scan the given python file(s) for all its imports. Then it fetches all package names (from the dependencies spec file). Finally, all imports are associated with their corresponding package name (requires the virtual environment for resolving). If a package does not have any imports associated, it will be considered to be unused.\n\n## Ambition and history\n\nThe idea of a package like this was born from having gotten security vulnerability\nreports about production dependencies (shipped into production) which turned out to not not\neven be in use.\n\nThe goal would be to be able to run this tool in CI, which will catch cases where the developer\nforgets to remove unused packages. A example of such a case could be when doing refactorings.\n\nThis can work well in tandem with flake8 or pylint, which can warn in CI about unused imports.\n\nNote: The creosote tool supports identifying both unused production dependencies and developer dependencies.\n\n## FAQ\n\n### Are requirements.txt files supported?\n\nYes, kind of. There is no way to tell which part of `requirements.txt` specifies production vs developer dependencies. Therefore, you have to break your `requirements.txt` file into e.g. `requirements-prod.txt` and `requirements-dev.txt` and use any of them as input.\n\nIf you are using [pip-tools](https://github.com/jazzband/pip-tools), you can provide a `*.in` file.\n\n### Can I scan for pyproject\'s dev-dependencies?\n\nYes! For `pyproject.toml`, just provide the `--dev` argument.\n\n### Can I use this as a GitHub Action?\n\nYes! See [.github/workflows/action.yaml](.github/workflows/action.yaml) for a working example.\n\n### What\'s with the name "creosote"?\n\nThis library has borrowed its name from the [Monty Python scene about Mr. Creosote](https://www.youtube.com/watch?v=aczPDGC3f8U).\n\n\n### Releasing\n\n1. Bump version with Poetry, e.g. `poetry version minor`.\n2. GitHub Action will run automatically on creating [a release](https://github.com/fredrikaverpil/creosote/releases).\n',
    'author': 'Fredrik Averpil',
    'author_email': 'fredrik.averpil@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fredrikaverpil/creosote',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.12',
}


setup(**setup_kwargs)
